"""
This module contains the data_structures used in py_search. In particular, it
contains the the :class:`Problem` class, which is used to represent the
different search problems, and the :class:`AnnotatedProblem` class, which wraps
around a specific problem and keeps track of the number of core method calls.

At a lower level this module also contains the :class:`Node` class, which is
used to represent a node in the search space.

Finally, the module contains the :class:`Fringe` class, and its instantiations
(:class:`FIFOQueue`, :class:`LIFOQueue`, and :class:`PrioritySet`). A Fringe is
used to structure the way a search space is explored.
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from collections import deque
from bisect import insort


class Problem(object):
    """
    The basic problem to solve. The main functions that must be defined include
    successors and goal_test. Some search techniques also require the
    random_successor and predecessors methods to be implemented.
    """
    def __init__(self, initial, parent=None, action=None, initial_cost=0,
                 extra=None):
        self.initial = Node(initial, parent, action, initial_cost, extra=extra)

    def node_value(self, node):
        """
        Returns the the value of the current node. This is the value being
        minimized by the search. By default the cost is used, but this
        function can be overloaded to include a heuristic.
        """
        return node.cost()

    def successors(self, node):
        """
        An iterator that yields all of the successors of the current node.
        """
        raise NotImplemented("No successors function implemented")

    def random_successor(self, node):
        """
        This method should return a single successor node. This is used
        by some of the local search / optimization techniques.
        """
        raise NotImplemented("No random successor implemented!")

    def random_node(self):
        """
        This method returns a random node in the search space. This
        is used by some of the local search / optimization techniques.
        """
        raise NotImplemented("No random node implemented!")

    def goal_test(self, node):
        """
        Returns true if a goal state is found. This is typically used not used
        by the local search / optimization techniques.
        """
        raise NotImplemented("No goal test function implemented")


class AnnotatedProblem(Problem):
    """
    A Problem class that wraps around another Problem and keeps stats on nodes
    expanded and goal tests performed.
    """
    def __init__(self, problem):
        self.problem = problem
        self.initial = problem.initial
        self.nodes_expanded = 0
        self.goal_tests = 0
        self.nodes_evaluated = 0

    def random_successor(self, node):
        """
        A wrapper for the random_successor method that keeps track of the
        number of nodes expanded.
        """
        self.nodes_expanded += 1
        return self.problem.random_successor(node)

    def random_node(self):
        """
        A wrapper for the random_node method.
        """
        return self.problem.random_node()

    def node_value(self, node):
        """
        A wraper for the node value method that keeps track of the number of
        times a node value was calculated.
        """
        self.nodes_evaluated += 1
        return self.problem.node_value(node)

    def successors(self, node):
        """
        A wrapper for the successor method that keeps track of the number of
        nodes expanded.
        """
        for s in self.problem.successors(node):
            self.nodes_expanded += 1
            yield s

    def goal_test(self, node):
        """
        A wrapper for the goal_test method that keeps track of the number of
        goal_tests performed.
        """
        self.goal_tests += 1
        return self.problem.goal_test(node)


class Node(object):
    """
    A class to represent a node in the search. This node stores state
    information, path to the state, cost of the node, depth of the node, and
    any extra information.

    :param state: the state at this node
    :type state: object for tree search and hashable object for graph search
    :param parent: the node from which the current node was generated
    :type parent: :class:`Node`
    :param action: the action performed to transition from parent to current.
    :type action: typically a string, but can be any object
    :param cost: the cost of reaching the current node
    :type cost: float
    :param extra: extra information to store in this node, typically used to
                  store non-hashable information about the state.
    :type extra: object
    """
    def __init__(self, state, parent=None, action=None, node_cost=0,
                 extra=None):
        self.state = state
        self.parent = parent
        self.action = action
        self.node_cost = node_cost
        self.extra = extra

    def depth(self):
        """
        Returns the depth of the current node. Uses a loop to compute depth
        (no tail recursion in python).
        """
        curr = self
        depth = 0
        while curr.parent is not None:
            curr = curr.parent
            depth += 1
        return depth

    def cost(self):
        """
        Returns the cost of the current node.
        """
        return self.node_cost

    def path(self):
        """
        Returns a path (tuple of actions) from the initial to current node.
        """
        actions = []
        current = self
        while current.parent:
            actions.append(current.action)
            current = current.parent
        actions.reverse()
        return tuple(actions)

    def __str__(self):
        return str(self.state) + str(self.extra)

    def __repr__(self):
        return repr(self.state)

    def __hash__(self):
        return hash(self.state)

    def __eq__(self, other):
        return isinstance(other, Node) and self.state == other.state

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.node_cost < other.node_cost


class Fringe(object):
    """
    A template for a fringe class. Used to control the strategy of different
    search approaches.
    """

    def push(self, node):
        """
        Adds one node to the collection.
        """
        raise NotImplemented("No push method")

    def extend(self, nodes):
        """
        Given an iterator (`nodes`) adds all the nodes to the collection.
        """
        for n in nodes:
            self.push(n)

    def pop(self):
        """
        Pops a node off the collection.
        """
        raise NotImplemented("No pop method")

    def __len__(self):
        raise NotImplemented("No len method")


class FIFOQueue(Fringe):
    """
    A first-in-first-out queue. Used to get breadth first search behavior.

    >>> fifo = FIFOQueue()
    >>> fifo.push(0)
    >>> fifo.push(1)
    >>> fifo.push(2)
    >>> print(fifo.pop())
    0
    >>> print(fifo.pop())
    1
    >>> print(fifo.pop())
    2
    """

    def __init__(self):
        self.nodes = deque()

    def push(self, node):
        self.nodes.append(node)

    def remove(self, node):
        for i in range(self.nodes.count(node)):
            self.nodes.remove(node)

    def pop(self):
        return self.nodes.popleft()

    def __len__(self):
        return len(self.nodes)


class LIFOQueue(FIFOQueue):
    """
    A last-in-first-out queue. Used to get depth first search behavior.

    >>> lifo = LIFOQueue()
    >>> lifo.push(0)
    >>> lifo.push(1)
    >>> lifo.push(2)
    >>> print(lifo.pop())
    2
    >>> print(lifo.pop())
    1
    >>> print(lifo.pop())
    0
    """

    def pop(self):
        return self.nodes.pop()


class PriorityQueue(Fringe):
    """
    A priority queue that sorts elements by their value. Always returns the
    minimum value item.  A :class:`PriorityQueue` accepts a node_value
    function, a cost_limit (nodes with a value greater than this limit will not
    be added) and a max_length parameter. If adding an item ever causes the
    size to exceed the max_length then the worst nodes are removed until the
    list is equal to max_length.

    >>> pq = PriorityQueue(node_value=lambda x: x, max_length=3)
    >>> pq.push(6)
    >>> pq.push(0)
    >>> pq.push(2)
    >>> pq.push(6)
    >>> pq.push(7)
    >>> print(len(pq))
    3
    >>> print(pq.pop())
    0
    >>> print(pq.pop())
    2
    >>> print(pq.pop())
    6

    :param node_value: The node evaluation function (defaults to
        ``lambda x: x.cost()``)
    :type node_value: a function with one parameter for node
    :param cost_limit: the maximum value for elements in the set, if an item
        exceeds this limit then it will not be added (defaults to
        ``float('inf'))``
    :type cost_limit: float
    :param max_length: The maximum length of the list (defaults to
        ``float('inf')``
    :type max_length: int or ``float('inf')``
    """
    def __init__(self, node_value=lambda x: x.cost(), cost_limit=float('inf'),
                 max_length=float('inf')):
        self.nodes = []
        self.max_length = max_length
        self.cost_limit = cost_limit
        self.node_value = node_value

    def clear(self):
        """
        Empties the list.
        """
        self.nodes = []

    def peek(self):
        """
        Returns the best node.
        """
        return self.nodes[-1][1]

    def peek_value(self):
        """
        Returns the value of the best node.
        """
        return -self.nodes[-1][0]

    def update_cost_limit(self, cost_limit):
        """
        Updates the cost limit and removes any nodes that violate the new
        limit.
        """
        self.cost_limit = cost_limit
        for i in range(len(self.nodes)):
            if self.nodes[i][0] <= self.cost_limit:
                self.nodes = self.nodes[i:]
                break

    def push(self, node):
        """
        Push a node into the priority queue. If the node exceeds the cost limit
        then it is not added. If the max_length is exceeded by
        adding the node, then the worst node is discarded from the set.
        """
        value = self.node_value(node)

        if value > self.cost_limit:
            return

        insort(self.nodes, (-value, node))

        if len(self.nodes) > self.max_length:
            val, node = self.nodes.pop(0)

    def pop(self):
        """
        Pop the best value from the priority queue.
        """
        val, node = self.nodes.pop()
        return node

    def __len__(self):
        return len(self.nodes)



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from random import choice

from py_search.base import Problem
from py_search.base import Node
from py_search.uninformed import depth_first_search
from py_search.uninformed import breadth_first_search
from py_search.uninformed import iterative_deepening_search
from py_search.informed import best_first_search
from py_search.informed import iterative_deepening_best_first_search
from py_search.informed import widening_beam_search
from py_search.utils import compare_searches


class EightPuzzle:
    """
    An eight puzzle class that can be used to test different search algorithms.
    When first created the puzzle is in the solved state.
    """

    def __init__(self):
        self.state = (0, 1, 2, 3, 4, 5, 6, 7, 8)

    def __hash__(self):
        return hash(self.state)

    def __eq__(self, other):
        if isinstance(other, EightPuzzle):
            return self.state == other.state
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "%i%i%i\n%i%i%i\n%i%i%i" % (self.state[0], self.state[1],
                                           self.state[2], self.state[3],
                                           self.state[4], self.state[5],
                                           self.state[6], self.state[7],
                                           self.state[8])

    def copy(self):
        """
        Makes a deep copy of an EightPuzzle object.
        """
        new = EightPuzzle()
        new.state = tuple([i for i in self.state])
        return new

    def randomize(self, num_shuffles):
        """
        Randomizes an EightPuzzle by executing a random action `num_suffles`
        times.
        """
        state = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        self.state = tuple(state)

        for i in range(num_shuffles):
            self.executeAction(choice([a for a in self.legalActions()]))

    def executeAction(self, action):
        """
        Executes an action to the EightPuzzle object.

        :param action: the action to execute
        :type action: "up", "left", "right", or "down"
        """
        zeroIndex = self.state.index(0)
        successor = list(self.state)

        if action == 'up':
            successor[zeroIndex] = successor[zeroIndex + 3]
            successor[zeroIndex + 3] = 0
        elif action == 'left':
            successor[zeroIndex] = successor[zeroIndex + 1]
            successor[zeroIndex + 1] = 0
        elif action == 'right':
            successor[zeroIndex] = successor[zeroIndex - 1]
            successor[zeroIndex - 1] = 0
        elif action == 'down':
            successor[zeroIndex] = successor[zeroIndex - 3]
            successor[zeroIndex - 3] = 0

        self.state = tuple(successor)

    def legalActions(self):
        """
        Returns an iterator to the legal actions that can be executed in the
        current state.
        """
        zeroIndex = self.state.index(0)

        if zeroIndex in set([0, 1, 2, 3, 4, 5]):
            yield "up"
        if zeroIndex in set([0, 3, 6, 1, 4, 7]):
            yield "left"
        if zeroIndex in set([2, 5, 8, 1, 4, 7]):
            yield "right"
        if zeroIndex in set([3, 4, 5, 6, 7, 8]):
            yield "down"


class EightPuzzleProblem(Problem):
    """
    This class wraps around an Eight Puzzle object and instantiates the
    successor and goal test functions necessary for conducting search.

    This class also implements an heuristic function which is used to compute
    the value for each successor as cost to node + heuristic estimate of
    distance to goal. This yield A* search when used with best first search or
    a more greedy variant when used with Beam Search.
    """

    def misplaced_tile_heuristic(self, state):
        """
        The misplaced tiles heuristic.
        """
        goal = EightPuzzle()
        h = 0
        for i, v in enumerate(state.state):
            if state.state[i] != goal.state[i]:
                h += 1
        return h

    def node_value(self, node):
        """
        The function used to compute the value of a node.
        """
        return node.cost() + self.misplaced_tile_heuristic(node.state)

    def successors(self, node):
        """
        Computes successors and computes the value of the node as cost +
        heuristic, which yields A* search when using best first search.
        """
        for action in node.state.legalActions():
            new_state = node.state.copy()
            new_state.executeAction(action)
            path_cost = node.cost() + 1
            yield Node(new_state, node, action, path_cost)

    def goal_test(self, node):
        """
        Check if the goal state has been reached.
        """
        goal = EightPuzzle()
        return node.state == goal


class NoHeuristic(EightPuzzleProblem):
    """
    A variation on the Eight Puzzle Problem that has a heuristic for 0. This
    yields something equivelent to dijkstra's algorithm when used with best
    first search and a more greedy variant when used with Beam Search.
    """

    def node_value(self, node):
        return node.cost()


if __name__ == "__main__":

    puzzle = EightPuzzle()
    puzzle.randomize(10)

    initial = puzzle
    print("Puzzle being solved:")
    print(puzzle)
    print()

    compare_searches(problems=[EightPuzzleProblem(initial)],
                     searches=[depth_first_search,
                               breadth_first_search,
                               iterative_deepening_search,
                               best_first_search,
                               iterative_deepening_best_first_search,
                               widening_beam_search])



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from itertools import combinations
from random import normalvariate
from random import shuffle
from random import randint

from munkres import Munkres

from py_search.base import Problem
from py_search.base import Node
from py_search.utils import compare_searches
from py_search.informed import best_first_search
from py_search.informed import beam_search
from py_search.optimization import branch_and_bound
from py_search.optimization import hill_climbing
from py_search.optimization import simulated_annealing
from py_search.optimization import local_beam_search


def random_matrix(n):
    """
    Generates an a list of list of floats (representing an n x n matrix) where
    the values have mean 0 and std 1.

    This is used as cost matrix for an assignment problem.
    """
    return [[normalvariate(0, 1) for j in range(n)] for i in range(n)]


class AssignmentProblem(Problem):
    """
    A tree search version of the assignment problem. Starts with an initially
    empty assignment and then incrementally builds the assignment up adding one
    assignment per expansion.
    """

    def min_cost_heuristic(self, node):
        """
        A huristic specifying the minimum cost that could be achieved for
        unassigned rows.
        """
        node.state
        costs, unassigned = node.extra

        empty_rows = [i for i, v in enumerate(node.state) if v is None]

        min_possible = 0
        for r in empty_rows:
            sub_c = [v for i, v in enumerate(costs[r]) if i in unassigned]
            min_possible += min(sub_c)

        return min_possible

    def node_value(self, node):
        """
        The value of a node is the combination of the node cost and the
        min_cost heuristic
        """
        return node.cost() + self.min_cost_heuristic(node)

    def successors(self, node):
        """
        An iterator that yields the sucessors of the provided node.
        """
        state = node.state
        costs, unassigned = node.extra

        for i, v in enumerate(state):
            if v is None:
                for u in unassigned:
                    new_state = tuple([u if i == j else k
                                       for j, k in enumerate(state)])
                    new_unassigned = tuple([k for k in unassigned if k != u])

                    c = node.cost() + costs[i][u]
                    yield Node(new_state, node, (i, u), c,
                               extra=(costs, new_unassigned))

    def goal_test(self, node):
        """
        A test of whether a complete assignment has been reached.
        """
        state = node.state
        return None not in state


class LocalAssignmentProblem(Problem):
    """
    This class represents a local search version of the assignment problem.
    I.e., a random state is generated to start the search and then neighbors of
    the state can be expanded in order to reduce the solution cost.
    """

    def node_value(self, node):
        """
        Returns a lower bound on the solution cost reachable from the given
        node (or its children)
        """
        return float('-inf')

    def random_successor(self, node):
        """
        A function that returns a random successor of the current node. This is
        used by the simulated annealing function, so it doesn't have to expand
        all successors.

        A successor is generated by randomly flipping a pair of row to column
        assignments.
        """
        costs = node.extra[0]

        p = [0, 0]
        p[0] = randint(0, len(node.state)-1)
        p[1] = p[0]
        while p[0] == p[1]:
            p[1] = randint(0, len(node.state)-1)

        new_cost = node.cost()
        new_cost -= costs[p[0]][node.state[p[0]]]
        new_cost -= costs[p[1]][node.state[p[1]]]
        new_cost += costs[p[0]][node.state[p[1]]]
        new_cost += costs[p[1]][node.state[p[0]]]

        state = list(node.state)
        temp = state[p[0]]
        state[p[0]] = state[p[1]]
        state[p[1]] = temp

        return Node(tuple(state), node, p, new_cost, extra=node.extra)

    def successors(self, node):
        """
        Generates successor states by flipping each pair of row to column
        assignments.
        """
        costs = node.extra[0]

        for p in combinations(node.state, 2):
            new_cost = node.cost()
            new_cost -= costs[p[0]][node.state[p[0]]]
            new_cost -= costs[p[1]][node.state[p[1]]]
            new_cost += costs[p[0]][node.state[p[1]]]
            new_cost += costs[p[1]][node.state[p[0]]]

            state = list(node.state)
            temp = state[p[0]]
            state[p[0]] = state[p[1]]
            state[p[1]] = temp

            yield Node(tuple(state), node, p, new_cost, extra=node.extra)

    def random_node(self):
        """
        Generates a node that has a random assignment.
        """
        state = random_assignment(len(self.initial.state))
        return Node(state, node_cost=cost(state, self.initial.extra[0]),
                    extra=self.initial.extra)

    def goal_test(self, node):
        return False


def random_assignment(n):
    """
    Returns a random valid assignment for an n x n matrix
    """
    state = list(range(n))
    shuffle(state)
    return tuple(state)


def cost(assignment, costs):
    """
    Given an assignemnt and a cost matrix, returns the cost of the
    assignment.
    """
    cost = 0.0
    for row, col in enumerate(assignment):
        cost += costs[row][col]
    return cost


def print_matrix(m):
    """
    Print a matrix
    """
    for row in m:
        print("\t".join(["%0.2f" % v for v in row]))


if __name__ == "__main__":

    n = 8
    costs = random_matrix(n)

    print()
    print("####################################################")
    print("Randomly generated square cost matrix (%i x %i)" % (n, n))
    print("####################################################")
    print_matrix(costs)

    print()
    print("####################################################")
    print("Optimial solution using Munkres/Hungarian Algorithm")
    print("####################################################")

    m = Munkres()
    indices = m.compute(costs)
    best = tuple([v[1] for v in indices])
    print("Munkres Solution:")
    print(best)
    print("Munkres Cost:")
    print(cost(best, costs))
    print()

    print("######################################")
    print("Local Search / Optimization Techniques")
    print("######################################")

    initial = random_assignment(n)
    problem = LocalAssignmentProblem(initial,
                                     initial_cost=cost(initial, costs),
                                     extra=(costs,))
    print("Initial Assignment (randomly generated):")
    print(initial)
    print("Initial Assignment Cost:")
    print(problem.initial.cost())
    print()

    def local_beam_width2(problem):
        return local_beam_search(problem, beam_width=2)

    def greedy_annealing(problem):
        num_neighbors = (n * (n-1)) // 2
        return simulated_annealing(problem, initial_temp=0,
                                   temp_length=num_neighbors)

    def annealing(problem):
        num_neighbors = (n * (n-1)) // 2
        return simulated_annealing(problem, initial_temp=1.5,
                                   temp_length=num_neighbors)

    def depth_limited_branch_and_bound(problem):
        return branch_and_bound(problem, depth_limit=4)

    compare_searches(problems=[problem],
                     searches=[depth_limited_branch_and_bound,
                               hill_climbing,
                               local_beam_width2,
                               greedy_annealing,
                               annealing])

    print()
    print("###########################")
    print("Informed Search Techniques")
    print("###########################")

    # TREE SEARCH APPROACH
    empty = tuple([None for i in range(len(costs))])
    unassigned = [i for i in range(len(costs))]

    new_costs = [[c - min(row) for c in row] for row in costs]
    min_c = [min([row[c] for row in costs]) for c, v in enumerate(costs[0])]
    new_costs = [[v - min_c[c] for c, v in enumerate(row)] for row in costs]

    tree_problem = AssignmentProblem(empty, extra=(costs, unassigned))

    def beam_width2(problem):
        return beam_search(problem, beam_width=2)

    print()
    compare_searches(problems=[tree_problem],
                     searches=[beam_width2,
                               best_first_search])



"""
This module includes the core search methods :func:`tree_search` and
`graph_search` and the primary uninformed search techniques:
:func:`depth_first_search`, :func:`breadth_first_search`, and
:func:`iterative_deepening_search`.
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from py_search.base import LIFOQueue
from py_search.base import FIFOQueue


def tree_search(problem, fringe, depth_limit=float('inf')):
    """
    Perform tree search (i.e., search where states might be duplicated) using
    the given fringe class.Returns an iterators to the solutions, so more than
    one solution can be found.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param fringe: The fringe class to use.
    :type fringe: :class:`fringe`
    :param depth_limit: A limit for the depth of the search tree. If set to
        float('inf'), then depth is unlimited.
    :type depth_limit: int or float('inf')
    """
    fringe.push(problem.initial)

    while len(fringe) > 0:
        node = fringe.pop()

        if problem.goal_test(node):
            yield node
        elif depth_limit == float('inf') or node.depth() < depth_limit:
            fringe.extend(problem.successors(node))


def graph_search(problem, fringe, depth_limit=float('inf')):
    """
    Perform graph search (i.e., no duplicate states) using the given fringe
    class. Returns an iterator to the solutions, so more than one solution can
    be found.

    Note that the closed list will allow re-expansion of nodes with a lower
    cost.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param fringe: The fringe class to use.
    :type fringe: :class:`fringe`
    :param depth_limit: A limit for the depth of the search tree. If set to
        float('inf'), then depth is unlimited.
    :type depth_limit: int or float('inf')
    """
    closed = {}
    fringe.push(problem.initial)
    closed[problem.initial] = problem.initial.cost()

    while len(fringe) > 0:
        node = fringe.pop()
        if problem.goal_test(node):
            yield node
        elif depth_limit == float('inf') or node.depth() < depth_limit:
            for s in problem.successors(node):
                if s not in closed or s.cost() < closed[s]:
                    fringe.push(s)
                    closed[s] = s.cost()


def depth_first_search(problem, depth_limit=float('inf'), search=graph_search):
    """
    A simple implementation of depth-first search using a LIFO queue.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param search: A search algorithm to use (defaults to graph_search).
    :type search: :func:`graph_search` or :func`tree_search`
    :param depth_limit: A limit for the depth of the search tree. If set to
        float('inf'), then depth is unlimited.
    :type depth_limit: int or float('inf')
    """
    for solution in search(problem, LIFOQueue(), depth_limit):
        yield solution


def breadth_first_search(problem, depth_limit=float('inf'),
                         search=graph_search):
    """
    A simple implementation of depth-first search using a FIFO queue.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param search: A search algorithm to use (defaults to graph_search).
    :type search: :func:`graph_search` or :func`tree_search`
    :param depth_limit: A limit for the depth of the search tree. If set to
        float('inf'), then depth is unlimited.
    :type depth_limit: int or float('inf')
    """
    for solution in search(problem, FIFOQueue(), depth_limit):
        yield solution


def iterative_deepening_search(problem, search=graph_search,
                               initial_depth_limit=0, depth_inc=1,
                               max_depth_limit=float('inf')):
    """
    An implementation of iterative deepening search. This search is basically
    depth-limited depth first up to the depth limit. If no solution is found at
    the current depth limit then the depth limit is increased by depth_inc and
    the depth-limited depth first search is restarted.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param search: A search algorithm to use (defaults to graph_search).
    :type search: :func:`graph_search` or :func`tree_search`
    :param initial_depth_limit: The initial depth limit for the search.
    :type initial_depth_limit: int or float('inf')
    :param depth_inc: The amount to increase the depth limit after failure.
    :type depth_inc: int
    :param max_depth_limit: The maximum depth limit (default value of
        `float('inf')`)
    :type max_depth_limit: int or float('inf')
    """
    depth_limit = initial_depth_limit
    while depth_limit < max_depth_limit:
        for solution in depth_first_search(problem, depth_limit, search):
            yield solution
        depth_limit += depth_inc



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from random import random
from random import shuffle
from random import choice

from py_search.base import Problem
from py_search.base import Node
from py_search.utils import compare_searches
from py_search.optimization import hill_climbing
from py_search.optimization import simulated_annealing
from py_search.optimization import branch_and_bound


def generate_graph(n, p):
    """
    Generates a random graph for graph partitioning. n specifies the number of
    nodes and p specifies the probability that any pair of nodes has a
    transition.
    """
    V = set([i for i in range(n)])
    E = []
    for n1 in range(n):
        for n2 in range(n1+1, n):
            if random() < p:
                E.append((n1, n2))

    return V, E


def random_partition(V):
    V = list(V)
    shuffle(V)
    return frozenset(V[:len(V)//2])


class LocalGraphPartitionProblem(Problem):
    """
    This class represents a local search version of the graph partition
    problem.  I.e., a random state is generated to start the search and then
    neighbors of the state can be expanded in order to reduce the solution
    cost.
    """
    def node_value(self, node):
        """
        This function is used by the branch_and_bound approach to determine
        whether successor nodes have the potential to be better than the
        current node. If this just returns the node cost, then the algorithm
        will explore nodes greedily.
        """
        return node.cost()

    def random_successor(self, node):
        """
        A function that returns a random successor of the current node. This is
        used by the simulated annealing function, so it doesn't have to expand
        all successors.

        A successor is generated by randomly flipping a pair of row to column
        assignments.
        """
        p = set(node.state)
        V, E = node.extra
        not_p = V - p

        pV = choice(list(p))
        not_pV = choice(list(not_p))

        p.remove(pV)
        p.add(not_pV)

        return Node(frozenset(p), node, node_cost=cutsize(E, p),
                    extra=node.extra)

    def successors(self, node):
        """
        Generates successor states by flipping each pair of row to column
        assignments.
        """
        p = node.state
        V, E = node.extra
        not_p = V - p

        for pV in p:
            for not_pV in not_p:
                new_p = set(p)
                new_p.remove(pV)
                new_p.add(not_pV)

                yield Node(frozenset(new_p), node, node_cost=cutsize(E, p),
                           extra=node.extra)

    def random_node(self):
        """
        Generates a node that has a random assignment.
        """
        V, E = self.initial.extra
        p = random_partition(V)
        return Node(p, node_cost=cutsize(E, p),
                    extra=self.initial.extra)

    def goal_test(self, node):
        """
        The search should never terminate early.
        """
        return False


def cutsize(E, p):
    cuts = [e for e in E if (e[0] in p and e[1] not in p) or
                            (e[0] not in p and e[1] in p)]
    return len(cuts)


if __name__ == "__main__":

    n = 30
    p = 20 / (n-1)
    print(n, p)
    V, E = generate_graph(n, p)
    initial = random_partition(V)
    cost = cutsize(E, initial)

    print("######################################")
    print("Local Search / Optimization Techniques")
    print("######################################")

    problem = LocalGraphPartitionProblem(initial, initial_cost=cost,
                                         extra=(V, E))

    print("Initial Assignment Cost:")
    print(cutsize(E, initial))
    print("Number of successors:")
    print(len(list(problem.successors(problem.initial))))
    print()

    def annealing(problem):
        size = (n * (n//2)) // 2
        return simulated_annealing(problem, initial_temp=5.5, temp_length=size)

    def greedy_annealing(problem):
        size = (n * (n//2)) // 2
        return simulated_annealing(problem, initial_temp=0, temp_length=size)

    def depth_limited_branch_and_bound(problem):
        return branch_and_bound(problem, depth_limit=100)

    compare_searches(problems=[problem],
                     searches=[hill_climbing, annealing, greedy_annealing,
                               branch_and_bound])



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from random import shuffle

from py_search.base import Problem
from py_search.base import Node
from py_search.uninformed import depth_first_search
from py_search.uninformed import breadth_first_search
from py_search.informed import best_first_search
from py_search.informed import iterative_deepening_best_first_search
from py_search.informed import beam_search
from py_search.optimization import hill_climbing
from py_search.optimization import local_beam_search
from py_search.optimization import simulated_annealing
from py_search.optimization import branch_and_bound
from py_search.utils import compare_searches


class nQueens:
    """
    An nQueens puzzle object
    """

    def __init__(self, n):
        self.n = n
        self.state = tuple([None for i in range(n)])

    def __hash__(self):
        return hash(self.state)

    def __eq__(self, other):
        if isinstance(other, nQueens):
            return self.state == other.state
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return str(self)

    def __str__(self):
        output = ""
        for c in self.state:
            if c is None:
                output += "|" + "|".join([" " for i in range(self.n)]) + "|\n"
            else:
                output += ("|" + "|".join([" " for i in range(c)] + ["Q"] +
                                          [" " for i in range(self.n-c-1)]) +
                           "|\n")
        return output

    def copy(self):
        """
        Makes a deep copy of an nQueens object.
        """
        new = nQueens(self.n)
        new.state = tuple([i for i in self.state])
        return new

    def randomize(self):
        """
        Randomizes an nQueens by shuffling a list of row to columns
        assignments.
        """
        state = [i for i in range(self.n)]
        shuffle(state)
        self.state = tuple(state)

    def num_conflicts(self):
        """
        Returns a count of the number of conflicts. First checks if there are
        column conflicts (row conflicts are impossible because of
        representation). Then checks for diagonals.
        """
        conflicts = 0
        for r1 in range(len(self.state)):
            c1 = self.state[r1]
            if c1 is None:
                continue
            for r2 in range(r1+1, len(self.state)):
                c2 = self.state[r2]
                if c2 is None:
                    continue
                if c1 == c2:
                    conflicts += 1
                if r2 - r1 == c2 - c1:
                    conflicts += 1
        return conflicts


class nQueensProblem(Problem):
    """
    A class that wraps around the nQueens object. This version of the problem
    starts with an empty board and then progressively adds queens.
    """
    def node_value(self, node):
        """
        The function used to compute the value of a node.
        """
        if node.state.num_conflicts() > 0:
            return float('inf')

        num_remaining = len(set(node.state.state) - set([None])) - node.state.n
        return node.depth() + num_remaining

    def successors(self, node):
        """
        Generate all possible next queen states.
        """
        free_columns = (set([i for i in range(node.state.n)]) -
                        set(node.state.state))

        for row, col in enumerate(node.state.state):
            if col is None:
                for nc in free_columns:
                    new_state = node.state.copy()
                    ns = [i for i in node.state.state]
                    ns[row] = nc
                    new_state.state = tuple(ns)
                    yield Node(new_state, node, ('add-queen', row, nc),
                               new_state.num_conflicts())

    def goal_test(self, node):
        """
        Check if the goal state (i.e., no queen conflicts) has been reached.
        """
        return (len(set(node.state.state) - set([None])) == node.state.n and
                node.state.num_conflicts() == 0)


class LocalnQueensProblem(Problem):
    """
    A class that wraps around the nQueens object. This version of the problem
    starts with an empty board and then progressively adds queens.
    """
    def successors(self, node):
        """
        Generate all permutations of rows.
        """
        for r1 in range(len(node.state.state)):
            c1 = node.state.state[r1]

            for r2 in range(r1+1, len(node.state.state)):
                c2 = node.state.state[r2]

                new_state = node.state.copy()
                ns = [i for i in node.state.state]
                ns[r1] = c2
                ns[r2] = c1
                new_state.state = tuple(ns)
                cost = new_state.num_conflicts()
                yield Node(new_state, node, ('swap', (r1, c1), (r2, c2)), cost)

    def random_node(self):
        nq_state = self.initial.state.copy()
        nq_state.randomize()
        cost = nq_state.num_conflicts()
        return Node(nq_state, None, None, cost)

    def random_successor(self, node):
        """
        Generate all permutations of rows.
        """
        rows = [i for i in range(node.state.n)]
        shuffle(rows)

        r1 = rows[0]
        c1 = node.state.state[r1]
        r2 = rows[1]
        c2 = node.state.state[r2]

        new_state = node.state.copy()
        ns = [i for i in node.state.state]
        ns[r1] = c2
        ns[r2] = c1
        new_state.state = tuple(ns)
        cost = new_state.num_conflicts()
        return Node(new_state, node, ('swap', (r1, c1), (r2, c2)), cost)

    def goal_test(self, node):
        """
        Check if the goal state (i.e., no queen conflicts) has been reached.
        """
        return node.cost() == 0


if __name__ == "__main__":

    print("###################")
    print("BACKTRACKING SEARCH")
    print("###################")

    initial = nQueens(5)
    print("Empty %i-Queens Problem" % initial.n)
    print(initial)
    print()

    compare_searches(problems=[nQueensProblem(initial)],
                     searches=[depth_first_search,
                               breadth_first_search,
                               best_first_search,
                               iterative_deepening_best_first_search,
                               beam_search])
    print()

    print("##########################")
    print("LOCAL SEARCH / OPTIMZATION")
    print("##########################")

    initial = nQueens(10)
    initial.randomize()
    cost = initial.num_conflicts()
    print("Random %i-Queens Problem" % initial.n)
    print(initial)
    print(cost)
    print()

    def beam2(problem):
        return local_beam_search(problem, beam_width=2)

    def steepest_hill(problem):
        return hill_climbing(problem)

    def annealing(problem):
        size = problem.initial.state.n
        n_neighbors = (size * (size-1)) // 2
        return simulated_annealing(problem, initial_temp=1.8,
                                   temp_length=n_neighbors)

    def greedy_annealing(problem):
        size = problem.initial.state.n
        n_neighbors = (size * (size-1)) // 2
        return simulated_annealing(problem, initial_temp=0,
                                   temp_length=n_neighbors)

    compare_searches(problems=[LocalnQueensProblem(initial,
                                                   initial_cost=cost)],
                     searches=[best_first_search,
                               branch_and_bound,
                               beam2,
                               steepest_hill,
                               greedy_annealing,
                               annealing])
    print()

    initial = nQueens(50)
    initial.randomize()
    cost = initial.num_conflicts()
    print("Random %i-Queens Problem" % initial.n)
    print(initial)
    print(cost)
    print()

    compare_searches(problems=[LocalnQueensProblem(initial,
                                                   initial_cost=cost)],
                     searches=[steepest_hill,
                               greedy_annealing,
                               annealing])



"""
This module includes the informed search techniques: :func:`best_first_search`
(i.e., A*), :func:`iterative_deepening_best_first_search` (i.e., IDA*),
:func:`beam_search`, and :func:`widening_beam_search`. 
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from py_search.base import PriorityQueue
from py_search.uninformed import graph_search
from py_search.uninformed import tree_search

def best_first_search(problem, search=graph_search, cost_limit=float('inf')):
    """
    Cost limited best-first search. By default the cost limit is set to
    `float('inf')`, so it is identical to traditional best-first search. This
    implementation uses a priority set (i.e., a sorted list without duplicates)
    to maintain the fringe.

    If the problem.node_value is the cost to reach the node plus an admissible
    heuristic estimate of the distance from the node to the goal, then this is the
    A* algorithm. 

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param search: A search algorithm to use (defaults to graph_search).
    :type search: :func:`graph_search` or :func`tree_search`
    :param cost_limit: The cost limit for the search (default = `float('inf')`)
    :type cost_limit: float
    """
    for solution in search(problem, PriorityQueue(node_value=problem.node_value,
                                                  cost_limit=cost_limit)):
        yield solution

def iterative_deepening_best_first_search(problem, search=graph_search,
                                          initial_cost_limit=0, cost_inc=1,
                                          max_cost_limit=float('inf')): 
    """
    A variant of iterative deepening that uses cost to determine the limit for
    expansion. When search fails, the cost limit is increased according to
    `cost_inc`. If the heuristic is admissible, then this is guranteed to find
    the best solution (similar to best first serach), but uses less memory.

    If the problem.node_value is the cost to reach the node plus an admissible
    heuristic estimate of the distance from the node to the goal, then this is the
    IDA* algorithm. 

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param search: A search algorithm to use (defaults to graph_search).
    :type search: :func:`graph_search` or :func`tree_search`
    :param initial_cost_limit: The initial cost limit for the search.
    :type initial_cost_limit: float
    :param cost_inc: The amount to increase the cost limit after failure.
    :type cost_inc: float
    :param max_cost_limit: The maximum cost limit (default value of `float('inf')`)
    :type max_cost_limit: float
    """
    cost_limit = initial_cost_limit
    while cost_limit < max_cost_limit:
        for solution in search(problem, PriorityQueue(cost_limit=cost_limit,
                                                      node_value=problem.node_value)):
            yield solution
        cost_limit += cost_inc

def beam_search(problem, beam_width=1, graph_search=True):
    """
    A variant of breadth-first search where all nodes in the fringe
    are expanded, but the resulting new fringe is limited to have length
    beam_width, where the nodes with the worst value are dropped. The default
    beam width is 1, which yields greedy best-first search (i.e., hill climbing).

    There are different ways to implement beam search, namely best-first
    beam search and breadth-first beam search. According to:

        Wilt, C. M., Thayer, J. T., & Ruml, W. (2010). A comparison of
        greedy search algorithms. In Third Annual Symposium on Combinatorial
        Search.

    breadth-first beam search almost always performs better. They find that
    allowing the search to re-expand duplicate nodes if they have a lower cost
    improves search performance. Thus, our implementation is a breadth-first
    beam search that re-expand duplicate nodes with lower cost.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param beam_width: The size of the beam (defaults to 1).
    :type beam_width: int
    :param graph_search: whether to use graph or tree search.
    :type graph_search: boolean
    """
    closed = {}
    fringe = PriorityQueue(node_value=problem.node_value)
    fringe.push(problem.initial)
    closed[problem.initial] = problem.initial.cost()

    while len(fringe) > 0:
        parents = []
        while len(fringe) > 0 and len(parents) < beam_width:
            parent = fringe.pop()
            if problem.goal_test(parent):
                yield parent
            parents.append(parent)
        fringe.clear()

        for node in parents:
            for s in problem.successors(node):
                if not graph_search:
                    fringe.push(s)
                elif s not in closed or s.cost() < closed[s]:
                    fringe.push(s)
                    closed[s] = s.cost()

def widening_beam_search(problem, initial_beam_width=1,
                               max_beam_width=1000, graph_search=True):
    """
    A variant of beam search that successively increase the beam width when
    search fails. This ensures that if a solution exists, and you're using graph
    search, and the solution space is finite, then that beam search will
    find it. 

    However, if you are looking for multiple solutions, then it might
    return duplicates as the width is increased. As the beam width increase, behavior
    is closer and closer to breadth-first search.

    :param problem: The problem to solve.
    :type problem: :class:`Problem`
    :param initial_beam_width: The initial size of the beam (defaults to 1).
    :type initial_beam_width: int
    :param max_beam_width: The maximum size of the beam (defaults to 1000).
    :type max_beam_width: int
    """
    beam_width = initial_beam_width
    while beam_width <= max_beam_width:
        for solution in beam_search(problem, beam_width=beam_width, graph_search=graph_search):
            yield solution
        beam_width += 1



"""
This module contains the local search / optimization techniques. Instead of
trying to find a goal state, these algorithms try to find the lowest cost
state.
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from math import exp
from math import log
from random import random

from py_search.base import PriorityQueue


def branch_and_bound(problem, graph_search=True, depth_limit=float('inf')):
    """
    An exhaustive optimization technique that is guranteed to give the best
    solution. In general the algorithm starts with some (potentially
    non-optimal) solution. Then it uses the cost of the current best solution
    to prune branches of the search that do not have any chance of being better
    than this solution (i.e., that have a node_value > current best cost).

    In this implementation, node_value should provide an admissible lower bound
    on the cost of solutions reachable from the provided node. If node_value is
    inadmissible, then optimality guarantees are lost.

    Also, if the search space is infinite and/or the node_value function
    provides too little guidance (e.g., node_value = float('-inf')), then
    the search might never terminate. To counter this, a depth_limit can be
    provided that stops expanding nodes after the provided depth. This will
    ensure the search is finite and guaranteed to terminate.

    Finally, the problem.goal_test function can be used to terminate search
    early if a good enough solution has been found. If goal_test(node) return
    True, then search is immediately terminated and the node is returned.

    Note, the current implementation uses best-first search via a priority
    queue data structure.

    :param problem: The problem to solve.
    :type problem: :class:`py_search.base.Problem`
    :param graph_search: Whether to use graph search (no duplicates) or tree
        search (duplicates)
    :type graph_search: Boolean
    """
    b = None
    bv = float('inf')

    fringe = PriorityQueue(node_value=problem.node_value)
    fringe.push(problem.initial)

    if graph_search:
        closed = set()
        closed.add(problem.initial)

    while len(fringe) > 0:
        pv = fringe.peek_value()

        if bv < pv:
            break

        node = fringe.pop()

        if problem.goal_test(node):
            yield node
            return

        if node.cost() < bv:
            b = node
            bv = node.cost()
            fringe.update_cost_limit(bv)

        if depth_limit == float('inf') or node.depth() < depth_limit:
            for s in problem.successors(node):
                if not graph_search:
                    fringe.push(s)
                elif s not in closed:
                    fringe.push(s)
                    closed.add(s)

    yield b


def hill_climbing(problem, random_restarts=0, max_sideways=0,
                  graph_search=True):
    """
    Probably the simplest optimization approach. It expands the list of
    neighbors and chooses the best neighbor (steepest descent hill climbing).

    Default configuration should yield similar behavior to
    :func:`local_beam_search` when it has a width of 1, but doesn't need to
    maintain alternatives, so might use slightly less memory (just stores the
    best node instead of limited length priority queue).

    If graph_search is true (the default), then a closed list is maintained.
    This is imporant for search spaces with platues because it keeps the
    algorithm from reexpanding neighbors with the same value and getting stuck
    in a loop.

    If random_restarts > 0, then search is restarted multiple times. This can
    be useful for getting out of local minimums.

    The problem.goal_test function can be used to terminate search early if a
    good enough solution has been found. If goal_test(node) return True, then
    search is immediately terminated and the node is returned.

    :param problem: The problem to solve.
    :type problem: :class:`py_search.base.Problem`
    :param random_restarts: The number of times to restart search. The
        initial state is used for the first search and subsequent starts begin
        at a random state.
    :type random_restarts: int
    :param max_sideways: Specifies the max number of contiguous sideways moves.
    :type max_sideways: int
    :param graph_search: Whether to use graph search (no duplicates) or tree
        search (duplicates)
    :type graph_search: Boolean
    """
    b = problem.initial
    bv = b.cost()

    if problem.goal_test(b):
        yield b

    if graph_search:
        closed = set()
        closed.add(problem.initial)

    c = b
    cv = bv

    while random_restarts >= 0:
        found_better = True
        sideways_moves = 0
        while found_better and sideways_moves <= max_sideways:
            found_better = False
            prev_cost = cv
            for s in problem.successors(c):
                if graph_search and s in closed:
                    continue
                elif graph_search:
                    closed.add(s)
                sv = s.cost()
                if sv <= bv:
                    b = s
                    bv = sv
                    if problem.goal_test(b):
                        yield b
                if sv <= cv:
                    c = s
                    cv = sv
                    found_better = True
            if cv == prev_cost:
                sideways_moves += 1
            else:
                sideways_moves = 0

        random_restarts -= 1
        if random_restarts >= 0:
            c = problem.random_node()
            while graph_search and c in closed:
                c = problem.random_node()
            cv = c.cost()

            if graph_search:
                closed.add(c)
            if cv <= bv:
                b = c
                bv = cv
                if problem.goal_test(b):
                    yield b

    yield b


def local_beam_search(problem, beam_width=1, max_sideways=0,
                      graph_search=True):
    """
    A variant of :func:`py_search.informed_search.beam_search` that can be
    applied to local search problems.  When the beam width of 1 this approach
    yields behavior similar to :func:`hill_climbing`.

    The problem.goal_test function can be used to terminate search early if a
    good enough solution has been found. If goal_test(node) return True, then
    search is immediately terminated and the node is returned.

    :param problem: The problem to solve.
    :type problem: :class:`py_search.base.Problem`
    :param beam_width: The size of the search beam.
    :type beam_width: int
    :param max_sideways: Specifies the max number of contiguous sideways moves.
    :type max_sideways: int
    :param graph_search: Whether to use graph search (no duplicates) or tree
        search (duplicates)
    :type graph_search: Boolean
    """
    b = None
    bv = float('inf')
    sideways_moves = 0

    fringe = PriorityQueue()
    fringe.push(problem.initial)

    while len(fringe) < beam_width:
        fringe.push(problem.random_node())

    if graph_search:
        closed = set()
        closed.add(problem.initial)

    while len(fringe) > 0 and sideways_moves <= max_sideways:
        pv = fringe.peek_value()

        if bv < pv:
            yield b

        if pv == bv:
            sideways_moves += 1
        else:
            sideways_moves = 0

        parents = []
        while len(fringe) > 0 and len(parents) < beam_width:
            parent = fringe.pop()
            parents.append(parent)
        fringe.clear()

        b = parents[0]
        bv = pv

        for node in parents:
            if problem.goal_test(node):
                yield node

            for s in problem.successors(node):
                if not graph_search:
                    fringe.push(s)
                elif s not in closed:
                    fringe.push(s)
                    closed.add(s)

    yield b


def simulated_annealing(problem, temp_factor=0.95, temp_length=None,
                        initial_temp=None, init_prob=0.4, min_accept=0.02,
                        min_change=1e-6, limit=float('inf')):
    """
    A more complicated optimization technique. At each iteration a random
    successor is expanded if it is better than the current node. If the random
    successor is not better than the current node, then it is expanded with
    some probability based on the temperature.

    Used the formulation of simulated annealing found in:
        Johnson, D. S., Aragon, C. R., McGeoch, L. A., & Schevon, C. (1989).
        Optimization by simulated annealing: an experimental evaluation; part
        I, graph partitioning. Operations research, 37(6), 865-892.

    Also, the problem.goal_test function can be used to terminate search early
    if a good enough solution has been found. If goal_test(node) return True,
    then search is immediately terminated and the node is returned.

    :param problem: The problem to solve.
    :type problem: :class:`py_search.base.Problem`
    :param temp_factor: The factor for geometric cooling, a value between 0 and
        1, but usually very close to 1.
    :type temp_factor: float
    :param temp_length: The number of nodes to expand at each temperature. If
        set to `None` (the default) then it is automatically chosen to be equal
        to the length of the successors list.
    :type temp_length: int
    :param initial_temp: The initial temperature for the annealing. The number
        is objective function specific. If set to None (the default), then
        a semi-random walk is used to select an initial
        temperature that will yield approx. init_prob acceptance rate for
        worse states.
    :type initial_temp: float or None
    :param min_accept: The fraction of states that must be accepted in
        temp_length iterations (taken from a single temperature) to not be
        frozen.  Every time this is not exceeded, the frozen counter is
        incremented until it hits 5. If a better state is found, then the
        frozen counter is reset to 0.
    :type min_accept: float between 0 and 1
    :param min_change: The amount of change that must be achieved in
        temp_length iterations (taken from a single temperature) to not be
        frozen.  Everytime this is not exceeded, the frozen counter is
        incremented until it hits 5. If a better state is found, then the
        frozen counter is reset to 0.
    :type min_change: float
    :param limit: The maximum number of iterations (random neighbors) to expand
        before stopping.
    :type limit: float
    """
    T = initial_temp
    b = problem.initial
    bv = b.cost()

    if problem.goal_test(b):
        yield b

    c = b
    cv = bv

    frozen = 0
    iterations = 0

    if temp_length is None:
        temp_length = len(list(problem.successors(c)))
        print("Temp length set equal to number of initial neighbors (%i)" %
              temp_length)

    if T is None:
        delta_sum = 0
        delta_count = 0

    while frozen < 5:
        acceptances = 0
        changes = 0

        for i in range(temp_length):
            iterations += 1
            s = problem.random_successor(c)
            sv = s.cost()

            if sv < bv:
                b = s
                bv = sv
                frozen = 0

            if iterations >= limit:
                yield b

            delta_e = sv - cv
            if T is None and delta_e > 0:
                delta_sum += delta_e
                delta_count += 1
            if ((delta_e <= 0 or (T is None and random() < 0.5) or
                 (T is not None and T > 1e-2 and random() < exp(-delta_e/T)))):
                acceptances += 1
                changes += delta_e
                c = s
                cv = sv

                if problem.goal_test(c):
                    yield c

        if T is None:
            if delta_count > 100:
                avg_delta = delta_sum / delta_count
                T = -avg_delta / log(init_prob)
                print("Initial temperature set to:",
                      "%0.3f (based on %i samples)" % (T, delta_count))
        else:
            T = temp_factor * T

        if (((acceptances / temp_length) < min_accept or
             abs(changes) < min_change)):
            frozen += 1

    yield b



"""
Utilities for the py_search library.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from tabulate import tabulate
import timeit

from py_search.base import AnnotatedProblem


def compare_searches(problems, searches):
    """
    A function for comparing different search algorithms on different problems.

    :param problems: problems to solve.
    :type problems: an iterator of problems (usually a list)
    :param searches: search algorithms to use.
    :type searches: an iterator of search functions (usually a list)
    """
    table = []

    for problem in problems:
        for search in searches:
            annotated_problem = AnnotatedProblem(problem)

            try:
                start_time = timeit.default_timer()
                sol = next(search(annotated_problem))
                elapsed = timeit.default_timer() - start_time
                cost = sol.cost()
            except StopIteration:
                value = 'Failed'
                elapsed = 'Failed'

            table.append([problem.__class__.__name__, search.__name__,
                          annotated_problem.goal_tests,
                          annotated_problem.nodes_expanded,
                          annotated_problem.nodes_evaluated, "%0.3f" % cost if
                          isinstance(cost, float) else cost,
                          "%0.4f" % elapsed if isinstance(elapsed, float) else
                         elapsed])

    print(tabulate(table, headers=['Problem', 'Search Alg', 'Goal Tests',
                                   'Nodes Expanded', 'Nodes Evaluated',
                                   'Solution Cost', 'Runtime'],
                   tablefmt="simple"))



import setuptools

setuptools.setup(
    name='py_search',
    version='1.2.1',
    author='Christopher J. MacLellan',
    author_email='maclellan.christopher@gmail.com',
    packages=setuptools.find_packages(),
    include_package_data=True,
    url='http://pypi.python.org/pypi/py_search/',
    license='LICENSE.txt',
    description=('A library of uninformed, informed, '
                 'and optimization search algorithms'),
    long_description=open('README.rst').read(),
    install_requires=['tabulate', 'munkres'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.3',
                 'Programming Language :: Python :: 2.4',
                 'Programming Language :: Python :: 2.5',
                 'Programming Language :: Python :: 2.6',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.0',
                 'Programming Language :: Python :: 3.1',
                 'Programming Language :: Python :: 3.2',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6'],
)
